/*  To prevent potential disclosure of the password to another user
    (when mistakenly passing wrong username), set PAM to call
    pam_gpgagent for the authentication management task only when
    it knows that the password matches.  */

//TODO: smazat nadbytečné "include" (např. podle mě nepotřebuji stdio.h)

#define _ISOC11_SOURCE
#define _GNU_SOURCE

#include <unistd.h>
#include <sys/wait.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <grp.h>
#include <pwd.h>

#define PAM_SM_AUTH
#define PAM_SM_ACCOUNT
#define PAM_SM_SESSION
#define PAM_SM_PASSWORD

#include <security/pam_modules.h>


#define LOCK_USE_SUFFIX "use"
#define LOCK_ACCESS_SUFFIX "access"
#define LOCK_DIR "/var/lock/pam_gpgagent"
#define MAKE_FILENAME(dir, name, suffix) ( make_filename (dir, name, suffix, sizeof (dir), sizeof (suffix)) )

struct sensitive_str_struct {
  size_t size;
  char str[];
};

typedef struct sensitive_str_struct t_sensitive_str;

static char const *const passwd_data_name = "gpgagent_passwd";
static char const *const use_fd_data_name = "gpgagent_use_fd";


#if 0
void
sensitive_str_wipe (void *data)
{
  volatile char *str;
  str = ((t_sensitive_str *)data)->str;

  while (*str)
    *str++ = '\0';

  munlock (data, ((t_sensitive_str *)data)->size);

  free (data);
}
#endif

void
sensitive_str_wipe (void *data, pam_handle_t *pamh)
{
  size_t size = ((t_sensitive_str *)data)->size;

  memset (data, 0, size);
  munlock (data, size);
  free (data);
}

void *
page_aligned_alloc (size_t *size)
{
  int alignment = getpagesize ();

  if (alignment <= 0)
    return NULL;

  /* Make sure ALIGNMENT is a multiple of 2
     (this is a requirement of aligned_alloc).  */
  if (alignment & 1)
    alignment <<= 1;

  /* Make *size the closest multiple of ALIGNMENT
     greater than or equal to original *size.  */
  *size += alignment - 1;
  *size -= *size % alignment;

  return aligned_alloc (alignment, *size);
}

char
sensitive_str_alloc (t_sensitive_str **dest, size_t size)
{
  struct rlimit rlp;

  if (size < 1)
    return -1;

  /* Check if core dumps were successfully prevented
     at module initialization.  */
  if (getrlimit (RLIMIT_CORE, &rlp))
    return -1;
  if ((rlp.rlim_cur != 0) || (rlp.rlim_max != 0))
    return -1;

  size += sizeof (t_sensitive_str);

  if (!(*dest = page_aligned_alloc (&size)))
    return -1;

  if (mlock (*dest, size))
    {
      free (*dest);
      return -1;
    }

  (*dest)->size = size;

  return 0;
}

/* Allocates memory to hold at most &size bytes of the string ORIG
   (including the zero byte at the end).
   Locks the allocated memory into RAM to prevent swapping it.
   Then copies the string ORIG, prepended with the actual size
   of the allocated region into thus prepared place.

   The address of the allocated region is put into *dest.

   It is safe to munlock (*dest, (*dest)->size) after wiping
   the memory, as both *dest and *size are guaranteed to be multiples
   of memory page size.

   Returns 0 on success, -1 on failure.  */
char
sensitive_str_copy (t_sensitive_str **dest, char const *orig, size_t size)
{
  if (size <= 0)
    return 0;

  size = strnlen (orig, size - 1);

  if (sensitive_str_alloc (dest, size + 1))
    return -1;

  memcpy ((*dest)->str, orig, size);
  (*dest)->str [size] = '\0';
  return 0;
}

static void
wipe_callback (pam_handle_t *pamh, void *data, int error_status)
{
  sensitive_str_wipe (data, pamh);
}

static void
free_callback (pam_handle_t *pamh, void *data, int error_status)
{
  free (data);
}


PAM_EXTERN int
pam_sm_authenticate (pam_handle_t *pamh, int flags, int argc,
                     char const **argv)
{
  void const *void_passwd;
  t_sensitive_str *sensitive_passwd;

  if (pam_get_item (pamh, PAM_AUTHTOK, &void_passwd) != PAM_SUCCESS)
    return PAM_AUTH_ERR;
  if (!void_passwd)
    return PAM_IGNORE;
  if (sensitive_str_copy (&sensitive_passwd, void_passwd, PAM_MAX_RESP_SIZE))
    return PAM_AUTH_ERR;
  if (pam_set_data (pamh, passwd_data_name, sensitive_passwd, &wipe_callback) != PAM_SUCCESS)
    {
      sensitive_str_wipe (sensitive_passwd, pamh);
      return PAM_AUTH_ERR;
    }
  return PAM_SUCCESS;
}

PAM_EXTERN int
pam_sm_setcred (pam_handle_t *pamh, int flags, int argc,
                char const **argv)
{
  return PAM_IGNORE;
}


PAM_EXTERN int
pam_sm_acct_mgmt (pam_handle_t *pamh, int flags, int argc,
                  char const **argv)
{
  return PAM_SERVICE_ERR;
}


static char
drop_root (char const *username)
{
  struct passwd *pw = getpwnam (username);
  if (pw)
    {
      uid_t uid = pw->pw_uid;
      gid_t gid = pw->pw_gid;
      if (initgroups (username, gid) ||
          /* as initgroups succeeded, we know that the effective uid is 0  */
          setgid (gid) ||
          setuid (uid))
        return -1;
      return 0;
    }
  return -1;
}

//FIXME: pipe() vytvari jadrem bufferovanou rouru - neni bezpecne pro citliva data (vyresi se pouzitim pipe2() s flagem O_DIRECT? - myslim, ze ne) - tak nejspis bych mel jadru duverovat :-)
static char
run_command (char const *filename, char const *const *argv, char const *data, char const *username, pam_handle_t *pamh)
{
  pid_t pid;
  int status;
  int link[2];

  if (pipe (link))
    return -1;

  pid = fork ();
  if (pid == 0)
    {
      void const *sensitive_passwd;

      /* In some configurations, fork() uses copy-on write to copy
         the heap. As exec() reclaims the heap, the password should not
         be copied at all in these systems, because between fork() and
         exec...() in the child process and between fork() and
         waitpid() in the parent process, we don't change the data in
         the pages containing the password.

         However, on systems without mm, fork() copies the heap,
         including the password.

         Note that "Memory locks are not inherited by a child created
         via fork(2)".

         For the case the system does not support mm, we zero
         the sensitive part of memory by call to pam_set_data,
         believing that it haven't been swapped after the fork().

         However, on systems with mm, this zeroing triggers the actual
         copying of the memory pages. In order not to expose those
         systems to the possibility that the copied passsword would be
         swapped, we lock this part of memory prior to zeoring it.  */
      if ((pam_get_data (pamh, passwd_data_name, &sensitive_passwd) == PAM_SUCCESS) &&
          (sensitive_passwd) &&
          (mlock (sensitive_passwd, ((t_sensitive_str const *)sensitive_passwd)->size)))
        _exit (EXIT_FAILURE);

      if ((pam_set_data (pamh, passwd_data_name, NULL, NULL) == PAM_SUCCESS) &&
          (!close (link[1])) &&
          (dup2 (link[0], STDIN_FILENO) != -1) &&
          (!close (link[0])) &&
          (!drop_root (username)))
        /* For backward compatibility reasons, the signature of
           execv does not define arguments as constant strings.
           POSIX however requires that the function will not change
           these strings, so we may cast safely.  */
        execv (filename, (char *const *)argv);
      _exit (EXIT_FAILURE);
    }
  else if (pid > 0)
    {
      if (close (link[0]))
        {
          close (link[1]);
          waitpid (pid, NULL, 0);
          return -1;
        }
      FILE *pwstream;
      char const *passwd;
      size_t len;
      ssize_t written;

      /* We use low-level IO instead of streams, because streams
         usually use buffers, which would mean that copy of
         the password would reside somewhere in memory where we could
         not wipe it nor prevent from swapping it.  */
      len = strlen (data);
      while (len -= (written = TEMP_FAILURE_RETRY (write (link[1], data, len))))
        {
          if (written == -1)
            {
              close (link[1]);
              waitpid (pid, NULL, 0);
              return -1;
            }
          data += written;
        }

      if (close (link[1]))
        {
          waitpid (pid, NULL, 0);
          return -1;
        }

      if ((waitpid (pid, &status, 0) != pid) || (status != EXIT_SUCCESS))
        return -1;
      return 0;
    }
  else
  {
    close (link[0]);
    close (link[1]);
    return -1;
  }
}

/* Returns 0 on success, -1 on failure.  */
static char
run_gpg_agent_if_not_running (char const *username, t_sensitive_str const *sensitive_passwd, pam_handle_t *pamh)
{
  static char const *const argv1[] = {"gpg-connect-agent", "--quiet", "/bye", NULL};
  static char const *const argv2[] = {"gpg-preset-passphrase", "--preset", "28124B8E3714AF8BA0C311D5E242FCD937814A58", NULL};
  static char const *const argv3[] = {"gpg-preset-passphrase", "--preset", "A20678D21F99BD5363F4C0AC6F7DA47ADEF23280", NULL};

  if (strcmp ("trosos", username))
    return 0;

  if (run_command ("/usr/bin/gpg-connect-agent", argv1, "", username, pamh) ||
      run_command ("/usr/libexec/gpg-preset-passphrase", argv2, sensitive_passwd->str, username, pamh) ||
      run_command ("/usr/libexec/gpg-preset-passphrase", argv3, sensitive_passwd->str, username, pamh))
    return -1;
  return 0;
}

/* Returns 0 on success, -1 on failure.  */
static char
terminate_gpg_agent_if_running (char const *username, pam_handle_t *pamh)
{
  static char const *const argv[] = {"gpg-connect-agent", "--quiet", "--no-autostart", "KILLAGENT", "/bye", NULL};

  if (strcmp ("trosos", username))
    return 0;

  return (run_command ("/usr/bin/gpg-connect-agent", argv, "", username, pamh));
}

static char *
make_filename (char const *dir, char const *name, char const *suffix, size_t const dirsize, size_t const sufsize)
{
  char *result;
  char *p;
  size_t namelen;

  namelen = strlen (name);
  result = malloc (namelen + dirsize + sufsize + 1);
  p = mempcpy (result, dir, dirsize - 1);
  *p++ = '/';
  p = mempcpy (p, name, namelen);
  *p++ = '.';
  mempcpy (p, suffix, sufsize);
  return result;
}

static int
access_open (char const *username)
{
  int fd;
  char *filename = MAKE_FILENAME (LOCK_DIR, username, LOCK_ACCESS_SUFFIX);

  fd = open (filename, O_WRONLY | O_CREAT, S_IRUSR | S_IWUSR);
  free (filename);
  return fd;
}

static int
use_open (char const *username)
{
  int fd;
  char *filename = MAKE_FILENAME (LOCK_DIR, username, LOCK_USE_SUFFIX);

  fd = open (filename, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR);
  free (filename);
  return fd;
}

/* Returns 0 on success, -1 on failure.  */
static int
unuse_gpg_agent (char const *username, int use_fd, pam_handle_t *pamh)
{
  int access_fd;
  struct flock write_lock = {
    .l_type = F_WRLCK,
    .l_whence = SEEK_SET,
    .l_start = 0,
    .l_len = 0,
  };
  struct flock no_lock = {
    .l_type = F_UNLCK,
    .l_whence = SEEK_SET,
    .l_start = 0,
    .l_len = 0,
  };

  if (fcntl (use_fd, F_SETLK, &no_lock) == -1)
    {
      close (use_fd);
      return -1;
    }

  access_fd = access_open (username);

  if (access_fd < 0)
    {
      close (use_fd);
      return -1;
    }

  if (fcntl (access_fd, F_SETLK, &write_lock) == -1)
    {
      if (errno == EAGAIN || errno == EACCES)
        {
          if (close (access_fd))
            {
              close (use_fd);
              return -1;
            }
          return close (use_fd);
        } else {
          close (access_fd);
          close (use_fd);
          return -1;
        }
    }

  if (fcntl (use_fd, F_GETLK, &write_lock) == -1)
    {
      fcntl (access_fd, F_SETLK, &no_lock);
      close (access_fd);
      close (use_fd);
      return -1;
    }

  if (write_lock.l_type != F_UNLCK)
    {
      if (fcntl (access_fd, F_SETLK, &no_lock) == -1)
        {
          close (access_fd);
          close (use_fd);
          return -1;
        }
      if (close (access_fd))
        {
          close (use_fd);
          return -1;
        }
      return close (use_fd);
    }

  if (close (use_fd))
    {
      terminate_gpg_agent_if_running (username, pamh);
      fcntl (access_fd, F_SETLK, &no_lock);
      close (access_fd);
      return -1;
    }

  if (terminate_gpg_agent_if_running (username, pamh))
    {
      fcntl (access_fd, F_SETLK, &no_lock);
      close (access_fd);
      return -1;
    }

  if (fcntl (access_fd, F_SETLK, &no_lock) == -1)
    {
      close (access_fd);
      return -1;
    }

  if (close (access_fd))
    return -1;

  return 0;
}

/* Returns file descriptor of the use lock file on success, -1 on failure.  */
static int
use_gpg_agent (char const *username, t_sensitive_str const *sensitive_passwd, pam_handle_t *pamh)
{
  int access_fd;
  int use_fd;
  struct flock write_lock = {
    .l_type = F_WRLCK,
    .l_whence = SEEK_SET,
    .l_start = 0,
    .l_len = 0,
  };
  struct flock read_lock = {
    .l_type = F_RDLCK,
    .l_whence = SEEK_SET,
    .l_start = 0,
    .l_len = 0,
  };
  struct flock no_lock = {
    .l_type = F_UNLCK,
    .l_whence = SEEK_SET,
    .l_start = 0,
    .l_len = 0,
  };

  access_fd = access_open (username);

  if (access_fd < 0)
    return -1;

  use_fd = use_open (username);

  if (use_fd < 0)
    {
      close (access_fd);
      return -1;
    }

  if (fcntl (access_fd, F_SETLKW, &write_lock) == -1)
    {
      close (use_fd);
      close (access_fd);
      return -1;
    }

  if (fcntl (use_fd, F_GETLK, &write_lock) == -1)
    {
      fcntl (access_fd, F_SETLK, &no_lock);
      close (use_fd);
      close (access_fd);
      return -1;
    }

  if (write_lock.l_type == F_UNLCK)
    {
      if (run_gpg_agent_if_not_running (username, sensitive_passwd, pamh))
        {
          fcntl (access_fd, F_SETLK, &no_lock);
          close (use_fd);
          close (access_fd);
          return -1;
        }
    }

  if (fcntl (use_fd, F_SETLK, &read_lock) == -1)
    {
      fcntl (access_fd, F_SETLK, &no_lock);
      close (access_fd);
      unuse_gpg_agent (username, use_fd, pamh);
      return -1;
    }

  if (fcntl (access_fd, F_SETLK, &no_lock) == -1)
    {
      fcntl (use_fd, F_SETLK, &no_lock);
      close (access_fd);
      unuse_gpg_agent (username, use_fd, pamh);
      return -1;
    }

  if (close (access_fd))
    {
      fcntl (use_fd, F_SETLK, &no_lock);
      unuse_gpg_agent (username, use_fd, pamh);
      return -1;
    }

  return use_fd;
}

PAM_EXTERN int
pam_sm_open_session (pam_handle_t *pamh, int flags, int argc,
                     char const **argv)
{
  void const *sensitive_passwd;
  void const *void_username;
  int *use_fd_p;

  if (pam_get_item (pamh, PAM_USER, &void_username) != PAM_SUCCESS)
    {
      pam_set_data (pamh, passwd_data_name, NULL, NULL);
      return PAM_SESSION_ERR;
    }
  if (!void_username)
    {
      pam_set_data (pamh, passwd_data_name, NULL, NULL);
      return PAM_SESSION_ERR;
    }

  switch (pam_get_data (pamh, passwd_data_name, &sensitive_passwd))
    {
      case PAM_SUCCESS:
        break;

      case PAM_NO_MODULE_DATA:
        pam_set_data (pamh, passwd_data_name, NULL, NULL);
        return PAM_IGNORE;

      default:
        pam_set_data (pamh, passwd_data_name, NULL, NULL);
        return PAM_SESSION_ERR;
    }

  use_fd_p = malloc (sizeof (*use_fd_p));
  if (!use_fd_p)
    {
      pam_set_data (pamh, passwd_data_name, NULL, NULL);
      return PAM_SESSION_ERR;
    }
  *use_fd_p = use_gpg_agent ((char *)void_username, (t_sensitive_str *)sensitive_passwd, pamh);
  if (*use_fd_p == -1)
    {
      free (use_fd_p);
      pam_set_data (pamh, passwd_data_name, NULL, NULL);
      return PAM_SESSION_ERR;
    }

  if (pam_set_data (pamh, passwd_data_name, NULL, NULL) != PAM_SUCCESS)
    {
      close (*use_fd_p);
      free (use_fd_p);
      return PAM_SESSION_ERR;
    }

  if (pam_set_data (pamh, use_fd_data_name, use_fd_p, &free_callback) != PAM_SUCCESS)
    {
      close (*use_fd_p);
      free (use_fd_p);
      return PAM_SESSION_ERR;
    }

  return PAM_SUCCESS;
}

PAM_EXTERN int
pam_sm_close_session (pam_handle_t *pamh, int flags, int argc,
                      const char **argv)
{
  void const *void_username;
  void const *void_use_fd_p;

  if (pam_get_item (pamh, PAM_USER, &void_username) != PAM_SUCCESS)
    return PAM_SESSION_ERR;
  if (!void_username)
    return PAM_SESSION_ERR;

  if (pam_get_data (pamh, use_fd_data_name, &void_use_fd_p) != PAM_SUCCESS)
    return PAM_SESSION_ERR;

  if (unuse_gpg_agent ((char *)void_username, *(int *)void_use_fd_p, pamh))
    {
      return PAM_SESSION_ERR;
    }

  return PAM_SUCCESS;
}


PAM_EXTERN int
pam_sm_chauthtok (pam_handle_t *pamh, int flags, int argc,
                  char const **argv)
{
  return PAM_SERVICE_ERR;
}

/* Prevent core dumps as we are going to
   store sensitive data in memory.
   Also, create the directory LOCK_DIR
   if it is not present yet.  */
void __attribute__ ((constructor))
start (void)
{
  struct rlimit rlp;

  if (getrlimit (RLIMIT_CORE, &rlp))
    return;

  rlp.rlim_cur = 0;
  rlp.rlim_max = 0;

  setrlimit (RLIMIT_CORE, &rlp);

  mkdir (LOCK_DIR, S_IRWXU);
}
